#!/usr/bin/python

# Copyright 2014 Keith Clark, Peter Robinson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys,os
import threading
import random
import math
import pedroclient
import time

IS_VERSION_2 = sys.version_info < (3, 0)
if IS_VERSION_2:
    import Queue
    import Tkinter as tk
    import tkFont
else:
    import queue as Queue
    import tkinter as tk
    from tkinter import font as tkFont

message_queue = Queue.Queue()

NUM_ROBOTS = 2

black = 0, 0, 0
grey = 100,100,100
red = 255, 0, 0
blue = 0, 0, 255

#HOME_ROOMS = [0,3,4]
HOME_ROOMS = [0,3,5]  

ROBOT_COLOURS = ['#0000DD', '#b000b0', '#f46747']

WIDTH = 740           # screen width
HEIGHT = 740          # screen height
TEXT_WIDTH = 60
ROBOT_RADIUS = 10

SQUARE_SIZE = 4

SPEED_FACTOR = 3         # forward speed
RATE_FACTOR = 2
SPEED_UP_FACTOR = 2       # Mult factor for speed change away from centre
SQ_SLOW_DOWN_DIST = 900  # square of dist to centre for speed change

WALL_WIDTH = 4

DOOR_WIDTH = 48
ROOM_SIZE = WIDTH//SQUARE_SIZE
SEE_DIST_SQ = (ROOM_SIZE * 0.65)**2
TEXT_OFFSET = 3*ROOM_SIZE//8

COL1 = '#AA0000'
COL2 = '#AAAA00'
COL3 = '#00AA00'
COL4 = '#00AAAA'

CENTRE_LINES = [r*ROOM_SIZE + ROOM_SIZE//2 for r in range(SQUARE_SIZE)]

DOOR_COLOUR_DICT = {"blue":"#0000FF",
                    "green":"#00AA00",
                    "red":"#FF0000",
                    "yellow":"#E0E000"}
                    
DOORS0 = {(0,1): "blue",
         (2,3): "blue",
         #(0,4): "green",
         (1,5): "green",
         (2,6): "green",
         (5,6): "blue",
         (9,10): "blue",
         (4,5): "yellow",
         (8,9): "yellow",
         (4,8): "red",
         (8,12): "green",
         (3,7): "red",
         (6,7): "yellow",
         (10,11): "green",
         (5,9): "red",
         (6,10): "red",
         (9,13): "green",
         (10,14): "yellow",
         (11,15): "red",
         (12,13): "blue",
         (14,15): "blue"}

DOOR_STATUS = { (r1, r2) : 1 for r1, r2 in DOORS0}

DOORS = {(r1,r2):DOORS0[(r1, r2)] for r1, r2 in DOORS0}
DOORS.update({(r2,r1):DOORS0[(r1, r2)] for r1, r2 in DOORS0})

EDGES = { (r1, DOORS[(r1, r2)]):r2 for (r1,r2) in DOORS}


ROOM_CENTRES = [(ROOM_SIZE*c + ROOM_SIZE//2, ROOM_SIZE*r + ROOM_SIZE//2) \
                for r in range(SQUARE_SIZE) for c in range(SQUARE_SIZE)]

FLOOR_LINES = [(ROOM_CENTRES[r1], ROOM_CENTRES[r2]) for r1,r2 in DOORS0]

CHARGERS = [12, 15]
CHARGER_RADIUS = 20
CENTRE_RADIUS = 2

CAMERA_ANGLE = 45   # the camera vision angle
HC_ANGLE = CAMERA_ANGLE//2
CAMERA_ARC_COLOUR = '#C8C800'


TURN_SPEED = 2
MOVE_SPEED = 2



def door_angle(room1, room2) :
    r1,c1 = room1//SQUARE_SIZE, room1 % SQUARE_SIZE
    r2,c2 = room2//SQUARE_SIZE, room2 % SQUARE_SIZE
    if r1 == r2:
        if c1 > c2:
            return 180
        else:
            return 0
    elif c1 == c2:
        if r1 > r2:
            return 270
        else:
            return 90
    else:
        return None

ANGLES_TO_ROOM_ROOM  = {(r1, r2):door_angle(r1,r2) \
                        for r1 in range(SQUARE_SIZE**2) \
                        for r2 in range(SQUARE_SIZE**2)}

ANGLES_TO_DOORS = {(r1, door_angle(r1,r2)): DOORS[(r1, r2)] \
                   for (r1, r2) in DOORS}

ANGLE_DELTA = 5

ADJACENT = [-1, 1, -SQUARE_SIZE, SQUARE_SIZE]

def centre_point(r1, r2):
    c1x, c1y = ROOM_CENTRES[r1]
    c2x, c2y = ROOM_CENTRES[r2]
    return ((c1x+c2x)//2, (c1y + c2y)//2)
    
DOOR_CENTRES = {(r1,r2):centre_point(r1, r2) for (r1, r2) in DOORS}

ANGLE_NESW = {0:"east", 90:"south", 180:"west", 270:"north"}

def angle2point(frompt, topt):
    fx,fy = frompt
    tx,ty = topt
    angle = 180*(math.atan2(ty-fy, tx-fx)/math.pi)
    if angle < 0:
        angle += 360
    return angle

def sign(x) :
    if x > 0:
        return 1
    elif x < 0:
        return -1
    else:
        return 0
    
def xy2room(x, y):
    return (int(y+0.5)//ROOM_SIZE)*SQUARE_SIZE + (int(x+0.5)//ROOM_SIZE)

def get_axis(dir):
    if dir < 0: dir += 360
    if dir < ANGLE_DELTA or dir > 360 - ANGLE_DELTA:
        return 0
    if 90-ANGLE_DELTA < dir < 90+ANGLE_DELTA:
        return 90
    if 180-ANGLE_DELTA < dir < 180+ANGLE_DELTA:
        return 180
    if 270-ANGLE_DELTA < dir < 270+ANGLE_DELTA:
        return 270
    return None

def get_reverse_angle(dir):
    return (dir+180) % 360

def get_heading(dir, turn):
    dir += 90
    if dir >= 360:
        dir -= 360
    if turn >= 0:
        return 10*(dir//10)
    else:
        d = 10*((dir+9)//10)
        if d == 360:
            return 0
        else:
            return d


def close_to_door_wall(r1, r2, x, y):
    cx, cy = DOOR_CENTRES[(r1, r2)]
    if abs(r1 - r2) == 1:
        return (abs(cx - x) < 15) and (abs(cy - y) < ROOM_SIZE/3)
    else:
        return (abs(cy - y) < 15) and (abs(cx - x) < ROOM_SIZE/3)
    
def is_same_dir(d1, d2):
    diff = abs(d1-d2)
    return diff < 5 or diff > 351


def action_in(a, actions):
    """return true iff actions contain an action with the same functor as a"""
    assert (a.get_type() == pedroclient.PObject.structtype), \
        str(a) + " is not a structure"
    a_val = a.functor.val
    for a1 in actions:
        assert (a1.get_type() == pedroclient.PObject.structtype), \
            str(a1) + " is not a structure"
        if a_val == a1.functor.val:
            return True
    return False

def get_start_stop_actions(old_actions, new_actions):    
    return ([a for a in new_actions if not a in old_actions],
            [a for a in old_actions if not action_in(a, new_actions)])


class GetNotifications(threading.Thread):
    def __init__( self, client, master):
        self.client = client
        self.master = master
        self.running = True
        threading.Thread.__init__(self)

    def run(self):
        """Run as thread - post an event when notifications arrive"""
    
        while (self.running):
            try:
                p2pmsg = self.client.get_term()[0]
                message_queue.put(p2pmsg)
                self.master.event_generate("<<pedro>>")
            except:
                break

    def stop(self):
        self.running = False
    



class Robot(object):
    def __init__(self, env, ID, x, y):
        self.x,self.y = x, y
        self.home_room = xy2room(x,y)
        self.dir = 270
        self.turning_speed = 0
        self.vx, self.vy = 0, 0
        self.id = ID
        self.env = env
        self.stopped = False
        self.queued_messages = []
        self.saved_see_doors = []
        self.saved_room_centre = False
        self.saved_see_room_centre = False
        self.saved_see_centre_close = False
        self.saved_pointing = None
        self.saved_in_doorway = None
        self.saved_pointing = None
        self.started = False
        self.path = []

    def set_path(self, path):
        points = [ROOM_CENTRES[r-1] for _, _, r in path]
        self.path = points
        
    def turn(self, sign):
        self.turning_speed = sign*TURN_SPEED

        
    def door_on_path(self):
        dx = int(100*round(math.cos(math.pi*self.dir/180)))
        dy = int(100*round(math.sin(math.pi*self.dir/180)))
        return DOORS.get((xy2room(self.x - dx, self.y - dy),
                          xy2room(self.x + dx, self.y + dy)))
        
        
    def get_room(self):
        return xy2room(self.x, self.y)

    def close_to_room_centre(self):
        cx, cy =  ROOM_CENTRES[self.get_room()]
        return (self.x - cx)**2 + (self.y - cy)**2 < 5
    
    def in_doorway(self):
        dx = int((float(self.x) / ROOM_SIZE) + 0.5) * ROOM_SIZE
        dy = int((float(self.y) / ROOM_SIZE) + 0.5) * ROOM_SIZE
        return abs(self.x - dx) < 5 or abs(self.y - dy) < 5
    
    def move(self, sign):
        dir = get_axis(self.dir)
        if dir is not None:
            self.dir = dir
        else:
            print("PATH ERROR !!!!! {} {} {} {}".format(self.dir, self.x, self.y, self.get_room()))
        self.vx = int(sign*MOVE_SPEED*round(math.cos(math.pi*self.dir/180)))
        self.vy = int(sign*MOVE_SPEED*round(math.sin(math.pi*self.dir/180)))
        
    def can_see_room_centre(self):
        angle = get_axis(self.dir)
        if angle is None:
            return False
        cx, cy = ROOM_CENTRES[self.get_room()]
        angle1 = get_axis(180*math.atan2((cy - self.y), (cx - self.x))/math.pi)
        dist = math.sqrt((self.x - cx)**2 + (self.y - cy)**2)
        return (dist > 0.1) and  (angle == angle1)

    def get_full_path(self):
        points = self.path[:]
        if points == []:
            return []
        ocx,ocy = points[-1]
        # if the last point is in different room and not in segment between
        # this rooms centre and next room centre
        if self.get_room() != xy2room(ocx,ocy):
            cx,cy = ROOM_CENTRES[self.get_room()]
            if (sign(cx - self.x) != sign(cx - ocx)) or \
               (sign(cy - self.y) != sign(cy - ocy)):
                points.append((cx,cy))
        points.append((self.x, self.y))
        return points


    def stop(self):
        self.turning_speed = 0
        self.vx, self.vy = 0, 0
        if self.close_to_room_centre():
            self.x, self.y = ROOM_CENTRES[self.get_room()]

    def update(self):
        if not self.started:
            return
        cx, cy =  ROOM_CENTRES[self.get_room()]
        if (self.x - cx)**2 + (self.y - cy)**2 < 900:
            vx = self.vx
            vy = self.vy
        else:
            vx = SPEED_UP_FACTOR*self.vx
            vy = SPEED_UP_FACTOR*self.vy
        self.x += vx
        self.y += vy
        self.dir += self.turning_speed
        if self.dir > 360:
            self.dir -= 360
        elif self.dir < 0:
            self.dir += 360
        if self.path and self.get_room() == xy2room(*self.path[-1]) and \
           self.close_to_room_centre():
            self.x, self.y =  ROOM_CENTRES[self.get_room()]
            self.path.pop()
        self.check_in_doorway()
        self.check_seeing_doors()
        self.check_pointing()
        self.check_centre()
        self.check_see_centre()
        self.send_messages()

    def check_centre(self):
        if self.close_to_room_centre():
            if not self.saved_room_centre:
                self.x, self.y = ROOM_CENTRES[self.get_room()]
                self.add_message('r_(at_room_centre)')
                self.saved_room_centre = True
        else:
            if self.saved_room_centre:
                self.add_message('f_(at_room_centre)')
                self.saved_room_centre = False

    def check_see_centre(self):
        angle = get_axis(self.dir)
        if angle is None:
            if self.saved_see_room_centre:
                self.add_message('f_(see_centre_ahead)')
                self.saved_see_room_centre = False
            if self.saved_see_centre_close:
                self.add_message('f_(see_centre_close)')
                self.saved_see_centre_close = False
            return

        if self.in_doorway():
            dx = int(10*round(math.cos(math.pi*self.dir/180)))
            dy = int(10*round(math.sin(math.pi*self.dir/180)))
            room = xy2room(self.x + dx, self.y + dy)
            is_in_doorway = True
        else:
            room = self.get_room()
            is_in_doorway = False
                
        cx, cy = ROOM_CENTRES[room]
        angle1 = get_axis(180*math.atan2((cy - self.y), (cx - self.x))/math.pi)
        dist = math.sqrt((self.x - cx)**2 + (self.y - cy)**2)
        if (dist > 0.1) and not is_in_doorway and(angle == angle1):
            # can see centre of current room
            if not self.saved_see_room_centre:
                self.add_message('r_(see_centre_ahead)')
                self.saved_see_room_centre = True
            if dist < 2*ROBOT_RADIUS + 10:
                if not self.saved_see_centre_close:
                    self.add_message('r_(see_centre_close)')
                    self.saved_see_centre_close = True
            else:
                if self.saved_see_centre_close:
                    self.add_message('f_(see_centre_close)')
                    self.saved_see_centre_close = False
        else:
            if self.saved_see_room_centre:
                self.add_message('f_(see_centre_ahead)')
                self.saved_see_room_centre = False
            if self.saved_see_centre_close:
                self.add_message('f_(see_centre_close)')
                self.saved_see_centre_close = False
                    
    def check_in_doorway(self):
        if self.in_doorway():
            door = self.door_on_path()
            if self.saved_in_doorway != door:
                self.add_message('r_(in_doorway({0}))'.format(door))
                self.saved_in_doorway = door
        elif not self.in_doorway() and  self.saved_in_doorway:
            self.add_message('f_(in_doorway({0}))'.\
                             format(self.saved_in_doorway))
            self.saved_in_doorway = None
            self.saved_see_doors = [] # ready to see doors in new room
        
    def check_seeing_doors(self):
        r = self.get_room()
        door_info = []
        for adj in ADJACENT:
            adjroom = r+adj
            doorcentre = DOOR_CENTRES.get((r, adjroom))
            if doorcentre:
                if (self.x - doorcentre[0])**2 + (self.y - doorcentre[1])**2 > \
                   SEE_DIST_SQ:
                    continue
                if r < adjroom:
                    r1, r2 = r,adjroom
                else:
                    r2, r1 = r,adjroom
                    
                door_status =  "open" if self.env.door_status[(r1,r2)] else "shut"
                door_colour = DOORS[(r,adjroom)]
                #door_dir = ANGLE_NESW[door_angle(r, adjroom)]
                door_info.append((door_colour, door_status))
        for info in self.saved_see_doors:
            if info not in door_info:
                self.add_message('f_(see_door({0}, {1}))'.format(*info))
        for info in door_info:
            if info not in self.saved_see_doors:
                self.add_message('r_(see_door({0}, {1}))'.format(*info))
        self.saved_see_doors = door_info
 
            
    def check_pointing(self):
        angle = get_heading(self.dir, self.turning_speed)
        if self.saved_pointing is None:
            self.add_message('r_(pointing({0}))'.format(angle))
            self.saved_pointing = angle
        elif angle != self.saved_pointing:
            self.add_message('f_(pointing({0}))'.format(self.saved_pointing))
            self.add_message('r_(pointing({0}))'.format(angle))
            self.saved_pointing = angle
                  
    def flipStopped(self):
        """
        if self.stopped:
            self.add_message("f_(simulation_stopped('$none_'))")
        else:
            self.add_message("r_(simulation_stopped('$none_'))")
        self.send_messages()
        """
        self.stopped = not self.stopped

    def add_message(self, msg):
        self.queued_messages.append(msg)

    def send_msg(self, msg):
        self.env.send_msg(self.id, msg)

    def shutdown(self, pedroclient):
        pedroclient.p2p('messages:robot{0}'.format(self.id), 'quiting')
        
    def send_messages(self):
        if self.queued_messages:
            print(self.queued_messages)
            message = '['
            for m in self.queued_messages:
                message += m + ','
            message = message[:-1]+']'
            self.send_msg(message)
            self.queued_messages = []

    def process_click(self, room, x, y):
        if room == self.get_room() and \
        (x - self.x)**2 + (y -self.y)**2 < 200 :
            if room in [12,15] and self.close_to_room_centre():
                self.add_message("f_(battery_status(low))")
                self.add_message("r_(battery_status(high))")
                self.send_messages()
            elif room == self.home_room:
                self.add_message("f_(battery_status(high))")
                self.add_message("r_(battery_status(low))")
                self.send_messages()

    def close_to_door(self, r1, r2):
        cx,cy = DOOR_CENTRES[(r1,r2)]
        return (abs(self.x - cx) < 40) and (abs(self.y - cy) < 40)
            
class Environment(object):
    def __init__(self, parent):
        self.parent = parent

        self.leftdown= False
        self.messages_to_send = None
        (r,c) = ROOM_CENTRES[HOME_ROOMS[0]]
        robot0 = Robot(self, 1, r,c)
        (r,c) = ROOM_CENTRES[HOME_ROOMS[1]]
        robot1 = Robot(self, 2, r,c)
        (r,c) = ROOM_CENTRES[HOME_ROOMS[2]]
        robot2 = Robot(self, 3, r,c)
        self.robots = [robot0, robot1, robot2]

        self.door_status = DOOR_STATUS
        self.task_actions = {}

    def shutdown(self, pedroclient):
        for r in self.robots:
            r.shutdown(pedroclient)
    def step(self):
        for r in self.robots:
            r.update()

    def flipStopped(self):
        for r in self.robots:
            r.flipStopped()

    def send_msg(self, ID, msg):
        self.parent.client.p2p("percepts:{0}{1}".format(self.sender_process, ID), msg)
        
    def process_msg(self, term):
        if term.type == pedroclient.PObject.structtype and \
           term.functor.val == "p2pmsg":
            msg = term.args[2]
            sender_process = term.args[1].args[0].args[1].val
            sender_task = term.args[1].args[0].args[0].val
            self.sender_process = sender_process[:-1]
            robotID = int(sender_process[-1])-1
            if str(msg) == "initialise_":
                self.robots[robotID].started = True
                self.robots[robotID].add_message('r_(battery_status(high))')
                self.robots[robotID].update()
                return
            elif  msg.get_type() == pedroclient.PObject.atomtype and \
                msg.val == '[]' :
                old_actions = self.task_actions.get(sender_task, [])
                self.task_actions[sender_task] = []
                self.process_stop_actions(old_actions, sender_task)
                return
            elif msg.get_type() == pedroclient.PObject.listtype:
                actions = msg.toList()
                old_actions = self.task_actions.get(sender_task, [])
                self.task_actions[sender_task] = actions
                (start_actions, stop_actions) \
                    = get_start_stop_actions(old_actions, actions)
                self.process_stop_actions(stop_actions, sender_task)
                self.process_start_actions(start_actions, sender_task)
                return
            else:
                    assert (False), "COMMAND: {}".format(str(pmsg))
        else:
            self.parent.text.configure(state="normal")
            self.parent.text.insert(tk.END, str(term)+'\n')
            self.parent.text.mark_set("insert",tk.END)
            self.parent.text.see(tk.END)
            self.parent.text.configure(state="disabled")
            if term.get_type() == pedroclient.PObject.structtype and \
                term.functor.val ==  'path':
                robotID = int(term.args[0].val[-1])-1
                print(str(term))
                if term.args[1].get_type() == pedroclient.PObject.listtype:
                    path = [(info.args[0].val, info.args[1].val,
                             info.args[2].val) \
                            for info in term.args[1].toList()]
                else:
                   path = []
                self.robots[robotID].set_path(path)
                return
            elif term.get_type() == pedroclient.PObject.structtype and \
                 term.functor.val ==  'no_path':
                robotID = int(term.args[0].val[-1])-1
                self.robots[robotID].set_path([])
                return
            
    def process_stop_actions(self, stop_actions, sender_task):
        robotID = int(sender_task[-1])-1
        for a in stop_actions:
            print("stop_action", sender_task, str(a))
            self.robots[robotID].stop()
            return

    def process_start_actions(self, start_actions, sender_task):
        robotID = int(sender_task[-1])-1
        #if start_actions:
        #    self.start_animate()
        for a in start_actions:
            print("start_action", sender_task, str(a))
            cmd_args = a.args
            cmd = a.functor.val
            if cmd == 'turn_at_centre':
                turndir = 1 if cmd_args[0].val == 'right' else -1
                self.robots[robotID].turn(turndir)
            elif cmd == 'follow_line_forwards':
                self.robots[robotID].move(1)
            elif cmd == 'follow_line_backwards':
                self.robots[robotID].move(-1)
            else:
                assert (False), "COMMAND: {}".format(str(a))

            
class RoomsApp(object):

    def __init__(self, master):
        self._master = master
        master.title("Rooms")
        master.protocol("WM_DELETE_WINDOW", self.onClose)
        #master.geometry("{0}x{1}".format(WIDTH+WALL_WIDTH+2+TEXT_WIDTH, HEIGHT+WALL_WIDTH))
        self.env = Environment(self)
        self.canvas = tk.Canvas(master, bg = 'white',
                                width = WIDTH+4, height=HEIGHT+4)
        self.canvas.pack(fill=tk.BOTH, expand=1, side=tk.LEFT)
        self.canvas.bind("<Button-1>", self.button_press_cb)
        self.canvas.bind("<Control-Button-1>", self.button_press_add_cb)
        self.canvas.bind("<Control-Shift-Button-1>", self.button_press_rem_cb)
        self.canvas.bind("<ButtonRelease-1>", self.button_release_cb)
        self.canvas.bind("<Button-2>", self.mid_button_press_cb)
        self.canvas.bind("<Button-3>", self.right_button_press_cb)
        self.canvas.bind("<B1-Motion>", self.motion_cb)
        self._master.bind("<<pedro>>", self.data_cb)
        scrollbar = tk.Scrollbar(master)
        scrollbar.pack(side=tk.RIGHT,fill=tk.Y)
        self.text = tk.Text(master,
                            yscrollcommand=scrollbar.set,
                            width=TEXT_WIDTH)
        self.text.pack(fill=tk.BOTH, expand=1, side=tk.LEFT)
        scrollbar.config(command=self.text.yview)
        self.canvas.bind_all("<Key>", self.onKeyPress)
        self.text.configure(state="disabled")
        self.text.bind("<1>", lambda event: self.text.focus_set())
        #self.font = tkFont.Font(family=FONT_NAME,size=FONT_SIZE,weight="bold")
        self.running = True
        self.client = pedroclient.PedroClient()
        self.client.register('rooms_env')
        self.client.subscribe("door_status(X,Y,Z)", "true")
        self.client.subscribe("release(X,Y)", "true")
        self.client.subscribe("path(X,Y,Z)", "true")
        self.client.subscribe("new_loc(X,Y,Z)", "true")
        self.client.subscribe("backed_up(X,Y,Z)", "true")
        self.client.subscribe("not_backed_up(X)", "true")
        self.client.subscribe("bid_to_move_to_centre(X, Y)", "true")
        self.client.subscribe("no_path(Rob)", "true")
        self.client.subscribe("join(Rob)", "true")
        self.client.subscribe("reserved_room(Rob, Rm)", "true")
        self.thread = GetNotifications(self.client, master)
        self.thread.setDaemon(True)
        self.thread.start()
        self.moving_bottle = None
        
        self.show()
        self._master.after(30, self.timer_cb)
        

    def onClose(self):
        self.env.shutdown(self.client)
        time.sleep(1)
        self._master.destroy()

    def show(self):
        self.canvas.delete(tk.ALL)
        self.redraw()


    def redraw(self):
        #font = wx.Font(FONT_SIZE, wx.MODERN, wx.NORMAL, wx.BOLD, 
        #               False, FONT_NAME)
        #dc.SetFont(font)
        #self.canvas.create_rectangle(0, 0, WIDTH+WALL_WIDTH, 
        #                             HEIGHT+WALL_WIDTH,
        #                             fill='white', outline='white')

        self.draw_room_nums()
        self.draw_walls()
        self.draw_floor_lines()
        self.draw_doors()
        self.draw_chargers()
        
        for robot in self.env.robots:
            self.drawRobotPath(robot)
        for robot in self.env.robots:
            self.drawRobot(robot)


    def draw_room_nums(self):
        for i, (cx, cy) in enumerate(ROOM_CENTRES):
            cx -= TEXT_OFFSET
            cy -= TEXT_OFFSET
            self.canvas.create_text(cx, cy, text=str(i+1)) 
        
    def draw_floor_lines(self):
        for cx,cy in ROOM_CENTRES:
            self.canvas.create_line((cx-2*ROBOT_RADIUS-5, cy),
                                    (cx+2*ROBOT_RADIUS+5, cy), fill='black')
            self.canvas.create_line((cx, cy-2*ROBOT_RADIUS-5),
                                    (cx, cy+2*ROBOT_RADIUS+5), fill='black')
                
        for p1,p2 in FLOOR_LINES:
            self.canvas.create_line(p1,p2, fill='black')

    def draw_walls(self):
        
        
        for r in range(SQUARE_SIZE+1):
            self.canvas.create_rectangle(r*ROOM_SIZE, 0,  
                                         r*ROOM_SIZE+WALL_WIDTH,
                                         HEIGHT+WALL_WIDTH,
                                         fill='black', outline='black')
        for r in range(SQUARE_SIZE+1):
            self.canvas.create_rectangle(0, r*ROOM_SIZE,   
                                         WIDTH+WALL_WIDTH,
                                         r*ROOM_SIZE + WALL_WIDTH,
                                         fill='black', outline='black')

        for (r1, r2) in DOORS0:
            r,c = r2 % SQUARE_SIZE, r2 // SQUARE_SIZE
            if r2 -r1 == 1:
                dt = c*ROOM_SIZE + ROOM_SIZE//2
                self.canvas.create_rectangle(r*ROOM_SIZE,
                                             dt - DOOR_WIDTH//2,  
                                             r*ROOM_SIZE+WALL_WIDTH,
                                             dt + DOOR_WIDTH//2,
                                             fill='white', outline='white')
            else:
                dt = r*ROOM_SIZE + ROOM_SIZE//2
                self.canvas.create_rectangle(dt - DOOR_WIDTH//2,
                                             c*ROOM_SIZE, 
                                             dt + DOOR_WIDTH//2,
                                             c*ROOM_SIZE+WALL_WIDTH,
                                             fill='white', outline='white')


                
    def draw_doors(self):
        for (r1, r2) in DOORS0:
            colour_name = DOORS0[(r1, r2)]
            rd = r2-r1
            colour = DOOR_COLOUR_DICT[colour_name]
            if rd == 1:
                # x-adjacent
                r,c = r2 % SQUARE_SIZE, r2 // SQUARE_SIZE
                dt = c*ROOM_SIZE + ROOM_SIZE//2
                if self.env.door_status[(r1, r2)] == 0:
                    self.canvas.create_rectangle(r*ROOM_SIZE-2,
                                                 dt - DOOR_WIDTH//2,  
                                                 r*ROOM_SIZE+WALL_WIDTH+2,
                                                 dt + DOOR_WIDTH//2,
                                                 fill=colour,
                                                 outline=colour)
                else:
                    self.canvas.create_rectangle(r*ROOM_SIZE-2,
                                                 dt + DOOR_WIDTH//2,  
                                                 r*ROOM_SIZE+WALL_WIDTH+2,
                                                 dt + 3*DOOR_WIDTH//2,
                                                 fill=colour, outline=colour)

                
            else:
                r,c = r2 % SQUARE_SIZE, r2 // SQUARE_SIZE
                dt = r*ROOM_SIZE + ROOM_SIZE//2
                if self.env.door_status[(r1, r2)] == 0:
                    self.canvas.create_rectangle(dt - DOOR_WIDTH//2,
                                                 c*ROOM_SIZE-2, 
                                                 dt + DOOR_WIDTH//2,
                                                 c*ROOM_SIZE+WALL_WIDTH+2,
                                                 fill=colour, outline=colour)
                else:
                    self.canvas.create_rectangle(dt + DOOR_WIDTH//2,
                                                 c*ROOM_SIZE-2, 
                                                 dt + 3*DOOR_WIDTH//2,
                                                 c*ROOM_SIZE+WALL_WIDTH+2,
                                                 fill=colour, outline=colour)

    def draw_chargers(self):
        for r in CHARGERS:
            r,c = r % SQUARE_SIZE, r // SQUARE_SIZE
            x = r*ROOM_SIZE + ROOM_SIZE//2
            y = c*ROOM_SIZE + ROOM_SIZE//2
            self.canvas.create_oval(x-CHARGER_RADIUS, y-CHARGER_RADIUS,
                                    x+CHARGER_RADIUS, y+CHARGER_RADIUS,
                                    fill='orange', 
                                    outline='orange')
            
    def drawRobotPath(self, robot):    
        path = robot.get_full_path()
        colour = ROBOT_COLOURS[robot.id-1]
        if path:
            self.canvas.create_line(*path, fill=colour,
                                    width=3,dash=(3,3))
    def drawRobot(self, robot):
        """Draw the robot"""
        x,y = robot.x, robot.y
        d = robot.dir
        colour = ROBOT_COLOURS[robot.id-1]
        self.canvas.create_oval(x-ROBOT_RADIUS, y-ROBOT_RADIUS,
                                x+ROBOT_RADIUS, y+ROBOT_RADIUS,
                                fill=colour, 
                                outline=colour)
        self.canvas.create_arc(x-ROBOT_RADIUS, y-ROBOT_RADIUS,
                               x+ROBOT_RADIUS, y+ROBOT_RADIUS,
                               start = (HC_ANGLE - d),
                               extent = -2*HC_ANGLE,
                               fill=CAMERA_ARC_COLOUR, 
                               outline=CAMERA_ARC_COLOUR)
        
    def onKeyPress(self, event):
        #self.text.tag_remove(tk.SEL, "1.0", tk.END)
        key = event.keysym
        if key != 'space':
            return 
        if self.running:
            self.canvas.create_text(330,530,text="STOPPED", fill='red')
        else:
            self.show()
            self._master.after(30, self.timer_cb)
        self.env.flipStopped()
        self.running = not self.running

    def timer_cb(self):
        if self.running:
            self.env.step()
            self.show()
            self._master.after(30, self.timer_cb)

    def data_cb(self, event):
        self.env.process_msg(message_queue.get())

    def button_press_cb(self, event):
        for r1,r2 in DOORS0:
            if close_to_door_wall(r1, r2, event.x, event.y):
                for robot in self.env.robots:
                    if robot.close_to_door(r1,r2):
                        return
                self.env.door_status[(r1,r2)] = 1 - \
                                                self.env.door_status[(r1,r2)]
                return
            
        room = xy2room(event.x, event.y)
        if room in [12, 15] or room in HOME_ROOMS:
            for robot in self.env.robots:
                robot.process_click(room, event.x, event.y)
            

        

    def button_press_add_cb(self, event):
        pass

    def button_press_rem_cb(self, event):
        pass

    def button_release_cb(self, event):
        pass

    def motion_cb(self, event):
        pass

    def mid_button_press_cb(self, event):
        pass

    def right_button_press_cb(self, event):
        pass



if __name__ == "__main__":
    root = tk.Tk()
    RoomsApp(root)
    root.mainloop()

