%% Copyright 2014 Keith Clark, Peter Robinson
%%
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.


?- compile_time_only(consult(op_decls)).

?- dynamic('__running__'/1, 0).
?- dynamic('__waiting__'/1, 0).
?- dynamic('__resources__'/2, 0).
?- dynamic('__time__'/1, 0).
?- dynamic('$delayed_tr_call'/1, 0).
?- dynamic('$agent_info'/2).
?- dynamic(overlapping_resources_/2).
?- dynamic(logger_name_/1).

?- dynamic('__percept_interface__'/1).

?- dynamic('$should_kill'/1).
?- dynamic('$percept_thread_running'/0).

?- dynamic('$can_run_handlers'/0).


%% If task atomic is not declared an we have multiple tasks then we use all_
%% as the collection of resources used.
all_resources_used_(R) :-
    '$task_atomic'(_), !,
    R = no.
all_resources_used_(yes).


%% upate the resources needed by Task to Res
'$update_resources'(Task , Res) :-
    '__resources__'(Task, Res), !.
'$update_resources'(Task, Res) :-
    retractall('__resources__'(Task, _)),
    assert('__resources__'(Task, Res)).


    
%%'__can_run__'(Task, Resources) :- ip_lookup('$processed', Processed),log_list(['__can_run__'(Task, Resources, Processed), nl_]) ,'__waiting__'(Task1), log_list([waiting___(Task1),nl_]), fail.
%% Check to see if Task, running TR, that needs Resources can become a
%% running task.
%% Note: this is called (as a negated test) in task atomic after first rule
%% is tested - failure of can_run causes the task to wait
'__can_run__'(Task, Resources) :-
    %% save away required resources
    '$update_resources'(Task, Resources),
    ip_lookup('$processed', Processed),
    %% Processed is the list of all running and waiting processes that have
    %% been processed. If Task is a non-running task then Processed includes
    %% all the waiting tasks earlier than Task in the wait queue (or Task
    %% is not yet added to the wait queue)
    '__can_run__aux'(Task, Resources, Processed).

%% If already running then can run
'__can_run__aux'(Task, _Resources, _) :-
    '__running__'(Task), !. %%,log_list([Task, " running",nl_]).
%% Task is not running or waiting - this occurs when Task was running
%% and the re-eval jumps out of the top-level task atomic - Task becomes
%% tentatively waiting and is to be rechecked when the tasks in the wait queue
%% have been checked if they can become running before this Task is checked
'__can_run__aux'(Task, _Resources, _) :-
    \+ '__waiting__'(Task), !,
    %%log_list([Task, " pending",nl_]),
    fail.
%% If Task is in the wait queue and its resources overlap with any running tasks
%% or tasks earlier in the wait queue (i.e. those in Pending)
'__can_run__aux'(Task, Resources, Processed) :-
    member(T, Processed),
    '__resources__'(T, TRes),
    overlap_res_(TRes, Resources), !,
    (
      '__waiting__'(Task)
    ->
      true
    ;
      assert('__waiting__'(Task))
    ), fail.
%% Task does not overlap with resources in running or waiting tasks earlier
%% in the wait queue so becomes running
'__can_run__aux'(Task, _Resources, _Processed) :-
    %%log_list([Task, " becomes running",nl_]),
    retractall('__waiting__'(Task)), 
    assert('__running__'(Task)).
 

%%overlap_res_(A, B) :- errornl(overlap_res_(A, B)), fail.
%% If all is used then resources always overlap
overlap_res_(all, _) :- !.
overlap_res_(_, all) :- !.
%% Overlap if there is a non-empty intersection
overlap_res_(Res1,Res2) :-
   member(R,Res1),
   member(R,Res2),!.
%% If the user defines overlapping_resources then this is also used to determine
%% if there is an overlap
overlap_res_(Res1,Res2) :-
    overlapping_resources_interface_(Res1, Res2), !.
overlap_res_(Res1,Res2) :-
    overlapping_resources_interface_(Res2, Res1), !.



%% get_delay_(Now, Times, Delay) is true
%% if Delay is the smallest element of the list Time bigger than now.
%% -1 if no such time.
get_delay_(Now, Times, Delay) :-
    get_min_time_(Times, Now, 1e20, Min),
    get_delay_aux_(Now, Min, Delay).

get_min_time_([], _, M, M).
get_min_time_([T|Rest], Now, InMin, Min) :-
    T >= Now, T < InMin, !,
    get_min_time_(Rest, Now, T, Min).
get_min_time_([_|Rest], Now, InMin, Min) :-
    get_min_time_(Rest, Now, InMin, Min).

get_delay_aux_(_Now, 1e20, -1) :- !.
get_delay_aux_(Now, Timeout, Delay) :- Delay is Timeout - Now.

%% This is called in non-task atomic TRs and for the first rule of a
%% task atomic when the Tr rule is either a refiring or an initial
%% firing

%%check_if_should_clean_up_(A, B) :- log_list([check_if_should_clean_up_(A,B), nl_]), fail.
check_if_should_clean_up_(yes, _) :- !.
check_if_should_clean_up_(_, none) :- !.
%% cleanup required
check_if_should_clean_up_(_, _) :-
    thread_symbol(Task),
    '$update_resources'(Task, []),
    (
      retract('__running__'(Task))
    ->
      %% the task was running - becomes tentatively waiting
      %% gets retried after waiting tasks get tried
      true
    ;
      '__waiting__'(Task)
    ->
      %% already waiting - do nothing
      true
    ;
      %% was tentatively waiting - becomes waiting
      assert('__waiting__'(Task))
    ).


%% When a TR rule has an attached action then they are collected during
%% evaluation and when evaluation is complete the list of actions are sent to
%% the action handler for processing
'$delay_tr_call'(G) :-
    assert('$delayed_tr_call'(G)).

'$send_action_messages' :-
    '$delayed_tr_call'(_), !,
    findall(G, retract('$delayed_tr_call'(G)), Actions),
    Actions ->> action_handler_.
'$send_action_messages'.



%% '$forward_on_eval_now'(Running, Waiting, Retry, Me, Processed)
%% Running is the list of running tasks to be re-evaluated
%% Waiting is the list of waiting tasks to be re-evaluated
%% Retry is the list of tentatively waiting tasks to be re-evaluated
%% Me is the current task
%% Processed is the list of processed (i.e. re-evaluated) tasks
%%'$forward_on_eval_now'(A, B, C, D,E) :- log_list(['$forward_on_eval_now'(A, B, C, D, E), nl_]), fail.
'$forward_on_eval_now'([], [], [], _, _) :-
    !,
    %% no more tasks to re-eval - restart handlers
    log_resources_,
    assert('$can_run_handlers').
'$forward_on_eval_now'([], [], [Head|Rest], Me, Processed) :-
    !,
    %% the next task was running - became waiting - re-re-eval
    eval_now([], [], Rest, Head, [Me|Processed], retry_waiting) ->> Head.
'$forward_on_eval_now'([], [Head|Rest], Retry, Me, Processed) :-
    '$not_running_or_waiting'(Me),
    !,
    %% tentatively waiting 
    assert('__waiting__'(Me)),
    %% the next task was waiting - add to retry list
    eval_now([], Rest, [Me|Retry], Head, Processed, no_timeout) ->> Head.
'$forward_on_eval_now'([], [Head|Rest], Retry, Me, Processed) :-
    !,
    %% the next task was waiting - add to processed
    eval_now([], Rest, Retry, Head, [Me|Processed], no_timeout) ->> Head.
'$forward_on_eval_now'([Head|Rest], Waiting, Retry, Me, Processed) :-
    '$not_running_or_waiting'(Me),
    !,
    assert('__waiting__'(Me)),
    %% this running task becomes waiting - add to the retry list
    eval_now(Rest, Waiting, [Me|Retry], Head, Processed, no_timeout) ->> Head.
'$forward_on_eval_now'([Head|Rest], Waiting, Retry, Me, Processed) :-
    %% task still running
    eval_now(Rest, Waiting, Retry, Head, [Me|Processed], no_timeout) ->> Head.

'$not_running_or_waiting'(Task) :-
    '__running__'(Task), !, fail.
'$not_running_or_waiting'(Task) :-
    '__waiting__'(Task), !, fail.
'$not_running_or_waiting'(_).

%% used in eval_no_resources
'$forward_on_no_resource_eval_now'([]) :-
    log_resources_,
    assert('$can_run_handlers').
'$forward_on_no_resource_eval_now'([First|Rest]) :-
    eval_now(Rest) ->> First.

%% eval_no_resources_ is used when there are no resources.
%% re-evaluation is triggered either by a timeout or by receiving an
%% eval_now message
%%eval_no_resources_(TermSeq, LastTerm,  Delay) :- errornl(eval_no_resources_(TermSeq, LastTerm, O Delay)),fail.
eval_no_resources_(TermSeq, LastTerm,  Delay) :- 
    Delay < 0, !,
    %% no delays - wait on eval_now message
    eval_now(Rest) <<= _,
    thread_atomic_goal(process_call_stack_(TermSeq, LastTerm,  NewLastTerm,
                                           NewDelay, no_timeout)),
    '$forward_on_no_resource_eval_now'(Rest),
    eval_no_resources_(TermSeq, NewLastTerm, NewDelay).
eval_no_resources_(TermSeq, LastTerm, Delay) :-
     ipc_recv(Message,_, [timeout(Delay)]),
     !,
     %% message received before timeout
     Message = eval_now(Rest),
    thread_atomic_goal(process_call_stack_(TermSeq, LastTerm,  NewLastTerm,
                                            NewDelay, no_timeout)),
    '$forward_on_no_resource_eval_now'(Rest),
    eval_no_resources_(TermSeq, NewLastTerm, NewDelay).
eval_no_resources_(TermSeq, LastTerm,  _Delay) :-
    %% delay ran out
    thread_atomic_goal(process_call_stack_(TermSeq, LastTerm, NewLastTerm,
                                            NewDelay, timeout)),
    !,
    eval_no_resources_(TermSeq, NewLastTerm,  NewDelay).

%% eval_ is used when there are resources and so potentially running
%% and waiting tasks
%% re-evaluation is triggered either by a timeout or by receiving an
%% eval_now message
%%eval_(A,B,C,D) :- errornl(eval_(A,B,C,D)),fail.
eval_(TermSeq, LastTerm, Delay) :-
    Delay < 0, !,
    eval_now(RunningTasks,WaitingTasks, ReTryTasks, Me, Processed, Kind) <<= _,
    %%log_list([eval_now(RunningTasks,WaitingTasks, ReTryTasks, Me, Processed)]),
    ip_set('$processed', Processed),
    %% it's time for this task thread to re-evaluate - do it atomically
    thread_atomic_goal(process_call_stack_(TermSeq, LastTerm,  NewLastTerm,
                                           NewDelay, Kind)),
    %% then make the recursive call
    %% pass on the message to the next task
    '$forward_on_eval_now'(RunningTasks,WaitingTasks,
                           ReTryTasks, Me, Processed),
    eval_(TermSeq, NewLastTerm, NewDelay).
eval_(TermSeq, LastTerm, Delay) :-
     %% there is a delay so wake up if delay expired
     ipc_recv(Message,_, [timeout(Delay)]),
     !,
     Message = eval_now(RunningTasks,WaitingTasks, ReTryTasks,
                        Me, Processed, Kind),
    %% woke up because of BS changes - so same as above
    ip_set('$processed', Processed),
    thread_atomic_goal(process_call_stack_(TermSeq, LastTerm, NewLastTerm,
                                           NewDelay, Kind)),
    %% pass on the message to the next task
    '$forward_on_eval_now'(RunningTasks,WaitingTasks,
                           ReTryTasks, Me, Processed),
    eval_(TermSeq, NewLastTerm, NewDelay).
eval_(TermSeq, LastTerm, _Delay) :-
    %% MinTime ran out
    %% do same as above except don't check if stamps have changed
    thread_atomic_goal(process_call_stack_(TermSeq, LastTerm,  NewLastTerm,
                                            NewDelay, timeout)),
    eval_(TermSeq, NewLastTerm, NewDelay).

%% process_call_stack_(TermSeq, LastTerm,
%%                     NewLastTerm,  NewDelay, Kind)
%% TermSeq is the call stack - the call will destructively update this
%% to the new call stack if the stack changes OW it remains the same.
%% LastTerm is the last term in the call stack - used to extract the
%% new delay time when no change is made to the call stack.
%% NewLastTerm is the new last term in the call stack - used as Last Term
%% in the recursive call to eval
%% NewDelay is the delay until the next time event (or -1 if none)
%% Kind is one of:
%% retry_waiting - when there is multiple tasks with resources - a running task
%% becomes waiting and then it is re-evaluated a second time - when this happens
%% Kind is set to retry_waiting
%% timeout - eval was woken because of a timeout.
%% no_timeout - neither of the above
%%process_call_stack_(A,B,C,D,E) :- errornl(process_call_stack_(A,B,C,D,E)),fail.
process_call_stack_(TermSeq, _, _, _, _Timedout) :-
    thread_symbol(Task),
    '$should_kill'(Task), !,
    %% task to be terminated
    retract('$should_kill'(Task)),
    do_action_([], TermSeq, -1),
    retractall('__resources__'(Task, _)),
    retractall('__waiting__'(Task)),
    retractall('__running__'(Task)),
    retractall(task(Task,_TR)),
    thread_exit.

process_call_stack_(TermSeq, _LastTerm, NewLastTerm, NewDelay, timeout) :-
    '$start_time'(T1),
    gettimeofday(T),
    Delta is T - T1,
    %ip_set('$current_time', T),
    open_string(write, Stream),
    thread_symbol(Task),
    writeL_(Stream, [Task, " woken because of timeout at ", Delta]),
    send_log_message_(Stream),
    TermSeq = [_|StrippedSeq],
    update_call_chain_aux_(StrippedSeq, TermSeq, NewLastTerm,
                           NewActions, NewDelay, T),
    !,
    do_action_(NewActions, TermSeq, NewDelay).
process_call_stack_(TermSeq, _LastTerm, NewLastTerm,
                     NewDelay, retry_waiting) :-
    gettimeofday(Now),
    TermSeq = [_|StrippedSeq],
    update_call_chain_aux_waiting_(StrippedSeq, TermSeq, NewLastTerm,
                                   NewActions, NewDelay, Now),
    !,
    do_action_(NewActions, TermSeq, NewDelay).
process_call_stack_(TermSeq, _LastTerm, NewLastTerm,
                     NewDelay, no_timeout) :-
    gettimeofday(Now),
    TermSeq = [_|StrippedSeq],
    update_call_chain_aux_(StrippedSeq, TermSeq, NewLastTerm,
                           NewActions, NewDelay, Now),
    !,
    do_action_(NewActions, TermSeq, NewDelay).
process_call_stack_(_TermSeq, LastTerm, LastTerm,
                     NewDelay, _) :-
    get_delay_from_stack_(LastTerm, NewDelay),
    open_string(write, Stream),
    thread_symbol(Task),
    writeL_(Stream, [Task, ": No change to the evaluation stack"]),
    send_log_message_(Stream).
    

%% update_call_chain_aux_(TermSeq, Trailing, LastTerm,
%%                        NewActions, Delay, Now)
%% TermSeq is the call stack, Trailing is such that Trailing = [_|TermSeq]
%% This is so the tail of Trailing (which becomes the new extension of the
%% call stack can be destructively updated

%%update_call_chain_aux_(A,B,C,D,E,F) :- errornl(update_call_chain_aux_(A,B,C,D,E,F)),fail.
update_call_chain_aux_([], _TermSeq, _, _Actions, _Delay, _Now) :-
    !,
    %% no changes
    fail.
update_call_chain_aux_([Term|_TermSeq], Trailing, NewLastTerm,
                       NewActions, Delay, Now) :-
    update_call_chain_aux_term_(Term, NewTermSeq, NewLastTerm, NewActions,
                               Now, Delay),
    %% Term on the call stack is updated to NewTermSeq and so the call
    %% stack from term on needs to be replaced by NewTermSeq
    setarg(2, Trailing, NewTermSeq),
    !.
update_call_chain_aux_([Term|TermSeq], [_|Trailing], LastTerm,
                       NewActions, Delay, Now) :-
    update_call_chain_aux_(TermSeq, Trailing, LastTerm,
                           NewActions, Delay, Now),
    '$set_to_unchanged'(Term).

%% This is purely for logging so we can see what part of the call stack is
%% continuing and where it changes to a new rule choice.
'$set_to_unchanged'(rule_choice(_, _, _, _, true)) :- !.
'$set_to_unchanged'(_).

%% This is the same as above but is used for re-re-evaluation - i.e.
%% when a re-evaluation causes the task to change from running to waiting.
%% After all other task call stacks are re-evaluated and we re-re-evaluate
%% in case the resources being used have changed in such a way that this
%% task can become running. Since the BS has not changed since the re-eval
%% then we don't need to re-evaluated every term innthe call stach - only the
%% last which will be a wait_choice term
%%update_call_chain_aux_waiting_(A,B,C,D,E,F) :- errornl(update_call_chain_aux_waiting_(A,B,C,D,E,F)),fail.
update_call_chain_aux_waiting_([], _TermSeq, _, _Actions, _Delay, _Now) :-
    !,
    %% no changes
    fail.
update_call_chain_aux_waiting_([Term|_TermSeq], Trailing,
                               NewLastTerm, NewActions, Delay, Now) :-
    Term = wait_choice(_TR),
    update_call_chain_aux_term_(Term, NewTermSeq, NewLastTerm, NewActions,
                               Now, Delay),
    setarg(2, Trailing, NewTermSeq),
    !.
update_call_chain_aux_waiting_([Term|TermSeq], [_|Trailing], LastTerm,
                       NewActions, Delay, Now) :-
    update_call_chain_aux_waiting_(TermSeq, Trailing, LastTerm,
                           NewActions, Delay, Now),
    '$set_to_unchanged'(Term).


%%update_call_chain_aux_term_(A,B,C,D,E,F) :- errornl(update_call_chain_aux_term_(A,B,C,D,E,F)),fail.
%% The entry is a timed sequence - has the timeout expired?
update_call_chain_aux_term_(timed_seq(Timeout, S, All, Time, InResources),
                            NewTermSeq, LastTerm, NewActions, Now, Delay) :-
    Timeout < Now,
    %% timeout expired - 
    %% get the next element in the timed sequence
    update_timed_seq_(S, All, S1), 
    S1 = [TR:T|_],
    update_call_chain_TS_(T, Now, S1, All, Time, OutTime,
                         InResources,  Term),
    update_call_chain_aux2_(TR, Term, RestTermSeq, LastTerm,
                           OutTime, Now, Delay,
                           NewActions, InResources),
    NewTermSeq = [Term|RestTermSeq].     
%% re-eval TR - was waiting
update_call_chain_aux_term_(wait_choice(TR), NewTermSeq, LastTerm,
                           NewActions, Now, Delay) :-
    %% The update is at a wait_choice
    tr_rule(TR, wait, Now, NewTermSeq, LastTerm, [], Delay,  NewActions, no).
%% re-eval TR
update_call_chain_aux_term_(rule_choice(TR,  RID, InTimes, ResourcesUsed, _),
                           NewTermSeq, LastTerm, NewActions, Now, Delay) :-
    tr_rule(TR, RID, Now, NewTermSeq, LastTerm, InTimes, Delay,
             NewActions, ResourcesUsed).


%%update_call_chain_TS_(A,B,C,D,E,F,G,H) :- errornl(update_call_chain_TS_(A,B,C,D,E,F,G,H)),fail.
%% at the last TS element with -1 as time - i.e. no timeout
update_call_chain_TS_(-1, _Now, S1, All, Time, OutTime,
                     InResources,  NewTerm) :-
    !,
    Time = OutTime,
    %% 1e20 is so far into the future that it's not a viable timeout
    NewTerm = timed_seq(1e20, S1, All, Time, InResources).
update_call_chain_TS_(T, Now, S1, All, Time, OutTime,
                     InResources, NewTerm) :-
    TSNow = Now + T,
    OutTime = [TSNow|Time],
    NewTerm = timed_seq(TSNow, S1, All, Time, InResources).

%%update_call_chain_aux2_(A,B,C,D,E,F,G,H,I) :- errornl(update_call_chain_aux2_(A,B,C,D,E,F,G,H,I)),fail.
% update_call_chain_aux2_(TR, _, _NewTermSeq, _, 
%                         _InTime, _Now, _Delay,
%                        _Actions, _InResources) :-
%       TR = pa([error]), !,
%       throw(tr_wait_repeat_error(TR)).
update_call_chain_aux2_(TR, Term, NewTermSeq, Term, 
                       InTime, Now,  OutDelay,
                        Actions, _InResources) :-
      TR = pa(Actions), !,
      %% the timed sequence element is a primitive action
      NewTermSeq = [],
      get_delay_(Now, InTime, OutDelay).

% update_call_chain_aux2_(TR, Term, NewTermSeq, Term, 
%                         InTime, Now, OutDelay,
%                         TR, _InResources) :-
%       TR = rpa(_Actions), !,
%       %% the timed sequence element is a primitive action
%       NewTermSeq = [],
%       get_delay_(Now, InTime, OutDelay).
update_call_chain_aux2_(TR, _, NewTermSeq, LastTerm, 
                       InTimes, Now, Delay,
                       Actions, InResources) :-
    gettimeofday(Now),
    %% otherwise a TR so update the call stack
    %% note if multi-tasking then the timed sequence must be inside
    %% a task atomic and therefore is running
    tr_rule(TR, none, Now, NewTermSeq, LastTerm,
            InTimes, Delay,  Actions, InResources).

%% get_delay_from_stack_(LastTerm, Delay)
%% LastTerm is the last term in the call stack and it contains all the
%% delays from all entries in the stack
%% Delay is the time to the next timeout in this delay info
%% Only used when there is no change to the call stack and so new delay
%% needs to be recomputed
%%get_delay_from_stack_(A,B) :- errornl(get_delay_from_stack_(A,B)),fail.
get_delay_from_stack_(rule_choice(_, rid(_, _, T), InTimes, _, _), Delay) :-
    nonvar(T),
    !,
    gettimeofday(Now),
    get_delay_(Now, [T|InTimes], Delay).
get_delay_from_stack_(rule_choice(_, _, InTimes, _, _), Delay) :-
    !,
    gettimeofday(Now),
    get_delay_(Now, InTimes, Delay).
      
get_delay_from_stack_(timed_seq(T, _, _, InTimes, _), Delay) :-
    !,
    gettimeofday(Now),
    get_delay_(Now, [T|InTimes], Delay).
%% this can only be because of a wait_choice and hence outside of a task atomic
%% and hence no delays
get_delay_from_stack_(_, -1).   


'$display_beliefs' :-
    '$start_time'(T1),
    '__time__'(T),
    Delta is T - T1,
    display_beliefs_(Delta).


%% update the list containing the remaining elements in timed sequence
update_timed_seq_(S, All, S1) :-
    S = [_], !, S1 = All.
update_timed_seq_(S, _All, S1) :-
    S = [_|S1].


do_action_(Actions, CallList, Delay) :-
    thread_symbol(Task),
    %% log the call stack and actions for Task
    log_call_stack_and_action_(Task, CallList, Actions,  Delay),
    %% Send off Actions to controller
    controls_(Actions),
    %% call attached ++ actions
    '$send_action_messages',
    !.  % to eliminate choice points


%% Start TR running
%% Exceptions are trapped and processed - currently a message is printed
%% and the task terminated
start_task_aux_(TR) :-
    catch(start_task_aux2_(TR), ErrorTerm, process_error_(ErrorTerm)).

    
start_task_aux2_(TR) :-
    '$task_atomic'(_),!,
    %% if task_atomic is defined then we expect multiple tasks and so we use
    %% eval_
    thread_symbol(Task),
    %% the IP '$processed' is used in '__can_run__' to see if
    %% this task can jump the wait queue - i.e. no resource overlaps
    %% with running and waiting tasks
    findall(T, ('__running__'(T) ; '__waiting__'(T)), Processed),
    ip_set('$processed', Processed),
    %% set task to waiting with no resources - used in start_task_atomic
    %% in conjunction with '__can_run__' to produce an initial call stack
    assert('__waiting__'(Task)),
    '$update_resources'(Task, []),
    thread_atomic_goal(start_task_atomic_(TR,  Delay,
                                         TermSeq, LastTerm)),
    log_resources_,
    %% enter eval loop
    eval_(TermSeq, LastTerm,  Delay).
start_task_aux2_(TR) :-
    %% no task_atomic so no resources  and no running/waiting
    %% so use eval_no_resources_
    thread_atomic_goal(start_task_atomic_(TR, Delay,
                                         TermSeq, LastTerm)),
    log_resources_,
    eval_no_resources_(TermSeq, LastTerm, Delay).

start_task_atomic_(TR,  Delay, TermSeq, LastTerm) :-
    gettimeofday(Now),
    %% reset time stamp for BS updates
    '$reset_time_fact',
    %% generate a new call stack
    tr_rule(TR, none, Now, TermSeq0, LastTerm, [], Delay, Action, no),
    %% A sentinal is added to the beginning of the call stack to allow
    %% efficient update of a tail of the call stack on re-evaluation
    TermSeq = [sentinal|TermSeq0],
    %% log info and process control (primitive) actions
    do_action_(Action, TermSeq, Delay).

process_error_(Error) :-
    thread_atomic_goal(process_error_aux_(Error)).

process_error_aux_(Error) :-
    thread_symbol(Task),
    display_task_error_(Error, Task),
    retractall('__running__'(Task)),
    retractall('__waiting__'(Task)),
    retractall('__resources__'(Task, _)),
    retract(task(Task,_)),
    send_robotic_message_interface_([task_ending(Task)]),
    thread_exit.
    
display_task_error_(error(no_matching_tr_rule, Prog, _, _), Task) :-
    !,
    write_term_list(stderr, ['No matching rule for ', Prog,
                             ' - task terminated.', nl]),
    log_error_(error(Task, no_matching_tr_rule, Prog)).

display_task_error_(ErrorTerm, _) :-
    write_term_list(stderr, ['Unhandled error ',
                             ErrorTerm,
                             ' - task terminated.', nl]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% logging

?- dynamic('$logging'/0).

%% all log messages are created by writing terms to a string stream
%% this predicate extracts the string representing the log message from Stream
%% and sends it to the logger
send_log_message_(Stream) :-
    logger_name_(Name), !,
    stream_to_string(Stream, String),
    String ->> Name.
%% if a logger had not been declared then simply close the stream
send_log_message_(Stream) :-
    close(Stream).

% log_command(T) :-
%     '$logging',!,
%     open_string(write, Stream),
%     write_term_list(Stream, T),
%     send_log_message_(Stream).
% log_command(_).

%% log messages from agents - called in handle_message_interface
log_message_(message(M, A)) :-
    '$logging', !,
    open_string(write, Stream),
    writeL_(Stream, ["Received message from ", A, " :", nl_, M]),
    send_log_message_(Stream).
log_message_(_).

% log_percepts(T) :-
%     '$logging',!,
%     open_string(write, Stream),
%     writeL_(Stream, T),
%     send_log_message_(Stream).
% log_percepts(_).

%% for logging task error messages
log_error_(T) :-
    %%do_log(errors),!,
    open_string(write, Stream),
    writeL_(Stream, [T]),
    send_log_message_(Stream).
%%log_error_(_).

%% For agents with resources logs the resources being used (or waited on)
log_resources_ :-
    %thread_symbol(Th), 
    '$logging',
    %%'__resources__'(Th,_ThRes),
    once(( '__running__'(_) ; '__waiting__'(_) )),
     !,
    open_string(write, Stream),
    thread_atomic_goal(
      (
        writeL_(Stream, ["Tasks using resources:", nl_]),
        forall('__running__'(Task),'$write_resources'(Task, Stream)),
        write(Stream, 'Tasks waiting on resources:\n'),
        forall('__waiting__'(Task), '$write_resources'(Task, Stream))
       )
     ),
    send_log_message_(Stream).
    
log_resources_.

'$write_resources'(Task, Stream) :-
    '__resources__'(Task, Res),
    (
     Res = [all__]
    ->
     writeTerm_(Stream, Task-all), nl(Stream)
    ;
     writeTerm_(Stream, Task-Res), nl(Stream)
    ).

%%log_call_stack_and_action_(Task, CallList, Actions, OldA, Controls, D) :- errornl(log_call_stack_and_action_(Task, CallList, Actions, OldA, Controls, D)),fail.


%% Log the call stack and actions
log_call_stack_and_action_(Task, [_|Chain], Actions, Delay) :-
    '$logging',!,
    open_string(write, Stream),
    '$start_time'(T1),
    writeL_(Stream, [Task, ": New call stack evaluation:", nl_]),
    %% print the call stack starting with an indent of 0 and bind D to
    %% the last indent
    log_action_chain_(Chain, 0, D, continue, Stream),
    D1 is D+1,
    writeL_(Stream, [sp_(D1), Actions, nl_]),
    (
      Delay = -1
    ->
      true
    ;
      gettimeofday(Now),
      DelayTill is Now - T1 + Delay,
      writeL_(Stream, [nl_, "Next call stack evaluation time event at ",
                      DelayTill,nl_]) 
    ),
    writeL_(Stream, [nl_, "Primitive actions for ",Task," are: ",Actions, nl_]),
    findall(A, '$delayed_tr_call'(A), QA),
    (
      QA = []
    ->
      true
    ;
      writeL_(Stream, ["Qulog actions: ", QA,nl_])
    ),
    send_log_message_(Stream).
log_call_stack_and_action_(_, _D, _, _).

log_manual_actions_(Controls) :-
    '$logging',!,
    open_string(write, Stream),
    writeL_(Stream, [nl_, "Primitive actions are: ",Controls, nl_]),
    send_log_message_(Stream).
log_manual_actions_( _Controls).

%%log_action_chain_(A,B,C,D,E) :- errornl(log_action_chain_(A,B,C,D,E)),fail.
log_action_chain_([], D, D, _, _).
log_action_chain_([no_matching_tr_rule(TR)|_Rest], D, D, _, Stream) :-
    !,
    writeL_(Stream, [sp_(D), TR, "- no matching rule", nl_]).
log_action_chain_([timed_seq(1e20, This, A, _, _)|Rest],
                 InD, OutD, Continue, Stream) :-
    !,
    length(This, ThisLen),
    length(A, ALen),
    Index is ALen - ThisLen + 1,
    writeL_(Stream, [sp_(InD), "["]),
    '$write_timed_seq_elems'(Stream, A),
    writeL_(Stream, ["] TS action ", Index,nl_]),
    MidD is InD + 1,
    log_action_chain_(Rest, MidD, OutD, Continue, Stream).
log_action_chain_([timed_seq((_ + Delay), This, A, _, _)|Rest],
                 InD, OutD, Continue, Stream) :-
    !,
    '$start_time'(T1),
    gettimeofday(T),
    NextTime is T - T1 + Delay,
    length(This, ThisLen),
    length(A, ALen),
    Index is ALen - ThisLen + 1,
    writeL_(Stream, [sp_(InD), "["]),
    '$write_timed_seq_elems'(Stream, A),
    writeL_(Stream, ["] TS action ", Index, " next TS action at ", NextTime,nl_]),
    MidD is InD + 1,
    log_action_chain_(Rest, MidD, OutD, Continue, Stream).
log_action_chain_([rule_choice(A, rid(I,_,_), _, _, New)|Rest],
                 InD, OutD, Continue, Stream) :-
    var(New), Continue = continue, !,
    %% var New means this is a new firing and if continue then this is the
    %% start of the new part of the call stack
    writeL_(Stream, [sp_(InD), "********************", nl_]),
    writeL_(Stream, [sp_(InD), A-I, " fired",nl_]),
    MidD is InD + 1,
    log_action_chain_(Rest, MidD, OutD, not_continued, Stream).
log_action_chain_([rule_choice(A, rid(I,_,_), _, _, New)|Rest],
                 InD, OutD, Continue, Stream) :-
    var(New),  !,
    %% a new firing but not the first one in the call stack
    writeL_(Stream, [sp_(InD), A-I, " fired",nl_]),
    MidD is InD + 1,
    log_action_chain_(Rest, MidD, OutD, Continue, Stream).
log_action_chain_([rule_choice(A, rid(I,_,_), _, _, _New)|Rest],
                 InD, OutD, Continue, Stream) :-
    !,
    %% continuing at this level
    writeL_(Stream, [sp_(InD), A-I, " continued",nl_]),
    MidD is InD + 1,
    log_action_chain_(Rest, MidD, OutD, Continue, Stream).
log_action_chain_([wait_choice(TR)|_Rest],
                 InD, InD, _Continue, Stream) :-
    !,
    %% call stack terminates in a wait
    writeL_(Stream, [sp_(InD), wait_choice(TR), nl_]).


'$write_timed_seq_elems'(_Stream, []).
'$write_timed_seq_elems'(Stream, [A]) :-
    !, '$write_timed_seq_elems_aux'(Stream, A).
'$write_timed_seq_elems'(Stream, [A|As]) :-
    '$write_timed_seq_elems_aux'(Stream, A),
    writeL_(Stream, [", "]),
    '$write_timed_seq_elems'(Stream, As).

'$write_timed_seq_elems_aux'(Stream, pa(As) : -1) :-
    !,
    write_a_tr_action_(Stream, As).
'$write_timed_seq_elems_aux'(Stream, pa(As) : T) :-
    !,
    write_a_tr_action_(Stream, As),
    writeL_(Stream, [" : ", T]).
'$write_timed_seq_elems_aux'(Stream, As : -1) :-
    !,
    writeTerm_(Stream, As).
'$write_timed_seq_elems_aux'(Stream, As : T) :-
    !,
    writeTerm_(Stream, As),
    writeL_(Stream, [" : ", T]).

    
log(all) :-
    assert('$logging').


logger_(Name) :-
    retract(logger_name_(OldName)) , !,
    writeL_(["Changing logger from ", OldName, " to ", Name, nl_]),
    assert(logger_name_(Name)).
logger_(Name) :-
    writeL_(["Starting logging to ", Name, nl_]),
    assert(logger_name_(Name)),
    assert('$logging').
unlog :-
    writeL_(["Stoping logging",nl_]),
    retractall('$logging'),
    retractall(logger_name_(_)).

%% Used to enter a list of beliefs that replace the existing beliefs.
%% Useful for debugging teleor programs as the user can check which TR
%% procedure guards are true for a given state of the belief store.
refresh_bs(_) :-
    forall(get_percepts_belief_(B), retract(B)),
    writeL_(stdout, ["Paste a list of beliefs from logger", nl_]),
    '$read_term_'(Beliefs),
    forall(member(B, Beliefs), remember(B)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Display beliefs

?- dynamic(task/2).

display_beliefs_(T) :-
    '$logging',
    !,
    open_string(write, Stream),
    thread_atomic_goal((display_all_beliefs_(T, Stream),
                        send_log_message_(Stream))).
display_beliefs_(_).

display_all_beliefs_(T, Stream) :-
    writeL_(Stream, [nl_,
                    "------ Current Percepts and Dynamic Beliefs at Time: ", T,
                    " ------",nl_]),
    findall(B, get_percepts_belief_(B), Beliefs),
    sort(Beliefs, SortedBeliefs, '$percept_order'),
    writeL_(Stream, [SortedBeliefs]).

%% used in sort above to list percepts in reasonable order
'$percept_order'(A, B) :-
    compound(A), compound(B), !,
    A =.. AArgs,
    B =.. BArgs,
    '$percept_order_args'(AArgs, BArgs).
'$percept_order'(A, B) :-
    A @=< B.

'$percept_order_args'([], _).
'$percept_order_args'([A|Arest], [B|BRest]) :-
    A = B, !,
    '$percept_order_args'(Arest, BRest).
'$percept_order_args'([A|_Arest], [B|_BRest]) :-
    '$percept_order'(A, B).


get_percepts_belief_(G) :-
    '$belief_type_info'(_, _, G),
    \+ '$nondisplay_belief'(G),
    call(G).

'$nondisplay_belief'(time_(_)).
'$nondisplay_belief'(task(_,_)).
'$nondisplay_belief'(running(_)).
'$nondisplay_belief'(waiting(_)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Percept management
%%

%% Keep asking the percept process to send initial percepts until
%% init percepts are sent - this does not read the percept message - this is
%% done by the caller
do_initialise_ :-
    '$agent_info'(_, Handle),
    same_handle(Handle, B@Host),
    repeat,
    to(initialise_, Handle),
    writeL_(["Waiting for initial percepts from ", Handle, nl_]),
    ipc_peek(_, H, [timeout(5)]),
    same_handle(H, B@Host), !,
    writeL_(["Received initial percepts from ", Handle, nl_]).

%% Embedded agents take care of percepts and control actions directly
%% rather via communicating with other agents.
%% This means that the percept handler needs to access sensors directly
%% and so polls the sensors every Reactivity seconds
start_embedded_percept_handler_(Reactivity) :-
    %% initialize the percept handler - default is to do nothing
    thread_atomic_goal(init_percept_handler_interface),
    %% get the initial sensors by calling the user defined poll_sensors
    assert('$can_run_handlers'),
    '$reset_time_fact',
    poll_sensors(Percepts),
    '$process_embedded_percepts_init'(Percepts),
    %% let other threads know handler is running
    assert('$percept_thread_running'),
    thread_yield,
    %% enter percept handler loop
    handle_embedded_percepts_(Reactivity).

%% For embedded agents: sleep then poll sensors and process them
handle_embedded_percepts_(Reactivity) :-
    repeat,
    thread_sleep(Reactivity),
    poll_sensors(P),
    thread_wait_on_goal('$can_run_handlers'),
    thread_atomic_goal('$process_embedded_percepts_init'(P)),
    fail.
    %handle_embedded_percepts_(Reactivity).

%% for initialization we don't want failure of '$process_embedded_percepts'
%% to cause failure of the caller
'$process_embedded_percepts_init'(P) :-
    '$process_embedded_percepts'(P), !.
'$process_embedded_percepts_init'(_).

%% For embedded agents we use the 'all' percept interface
'$process_embedded_percepts'(Percepts) :-
    '$log_process_percepts'(Percepts),
    '__time__'(BSTime),
    %% any remember/forget will cause __time__ to increase as
    %% this is the timestamp on BS updates
    handle_percepts_interface_(Percepts),
    '__time__'(NewBSTime),
    NewBSTime > BSTime,
    %% timestamp increased so need to check tasks   
    task(_, _),
    %% log new BS
    '$display_beliefs',
    %% handlers wait on '$can_run_handlers and so retracting it
    %% will block handlers from running
    retractall('$can_run_handlers'),
    %% there is at least one task and so all tasks need re-eval
    '$eval_tasks'.

'$eval_tasks' :-
    '$task_atomic'(_),!,  
    %% there is at least one task atomic and so there are running/waiting tasks
    findall(RTask, '__running__'(RTask), RunningTasks),
    findall(WTask, '__waiting__'(WTask), WaitingTasks),
    %% Send message to first task
    %% NOTE: running tasks are processed before waiting tasks
    '$send_eval_now_message'(RunningTasks, WaitingTasks).
'$eval_tasks' :-
    % no resources
    findall(Task, task(Task, _), Tasks),
    Tasks = [First|Rest],
    eval_now(Rest) ->> First.

%%'$send_eval_now_message'(A,B) :- log_list(['$send_eval_now_message'(A,B),nl_]),fail.
'$send_eval_now_message'([], [FirstTask|Rest]) :- !,
    %%log_list([eval_now([], Rest, [], FirstTask, []), " ->> ", FirstTask, nl_]),
    eval_now([], Rest, [], FirstTask, [], no_timeout) ->> FirstTask.
'$send_eval_now_message'([FirstTask|Rest], Waiting) :-
    %%log_list([eval_now(Rest, Waiting, [], FirstTask, []), " ->> ", FirstTask, nl_]),
    eval_now(Rest, Waiting, [], FirstTask, [], no_timeout) ->> FirstTask.

%% Start the percepts handling thread
start_percept_handler_ :-
    %% user defined interface - mostly not needed but the user
    %% can define if, for example, the percepts comms channel needs to be set up
    %% EG ROS interface
    thread_atomic_goal(init_percept_handler_interface),
    (
     current_predicate(handle_percepts_interface_/1)
    ->
     true
    ;
     '$assert_percepts_interface'(updates)
    ),
    %% This is used for managing if handler threads or eval threads should run
    %% When asserted percept, message and TR action threads can run.
    assert('$can_run_handlers'),
    '$reset_time_fact',
    %% Send an initialise_ message to "robot" and wait for a response
    %% to appear in the percept thread message buffer (using peek)
    do_initialise_,
    '$get_percepts'(P),
    '__percept_interface__'(Interface),
    %% process the read in percepts using the user declared interface
    %% (all or updates)
    '$process_percepts_init'(Interface, P),
    %% Signal to the forking thread that the percept handler is ready
    assert('$percept_thread_running'),
    %% Given up its timeslice so othere threads can run
    thread_yield,
    %% Enter percept handler loop
    handle_percepts_(Interface).       

%% The percept handler is a repeat fail loop rather than a recursive loop
%% to avoid garbage collection
handle_percepts_(Interface) :-
    repeat,
    '$get_percepts'(P),
    %% there are percepts to process - wait until task re-eval finished 
    thread_wait_on_goal('$can_run_handlers'),
    thread_atomic_goal('$process_percepts'(Interface, P)),
    fail.
   %% handle_percepts(Interface).

%% for initialization we don't want failure of '$process_percepts'
%% to cause failure of the caller
'$process_percepts_init'(Interface, P) :-
    '$process_percepts'(Interface, P),
    !.
'$process_percepts_init'(_, _).

'$process_percepts'(Interface, P) :-
    %% there might have been more percepts while waiting - get the rest
    %% and filter out invalid percepts terms
    '$get_all_filtered_percepts'(Interface, P, GPs),
    '$log_process_percepts'(GPs),
    '__time__'(BSTime),
    %% process the percepts (BS updates)
    handle_percepts_interface_(GPs),
    '__time__'(NewBSTime),
    NewBSTime > BSTime,
    %% timestamp increased so need to check tasks   
    task(_, _),
    %% log new BS
    '$display_beliefs',
    %% handlers wait on '$can_run_handlers and so retracting it
    %% will block handlers from running
    retractall('$can_run_handlers'),
    %% there is at least one task and so all tasks need re-eval
    '$eval_tasks', !.

%% get perepts list and filter out any terms that are not ground percept terms
%% using all interface
'$get_all_filtered_percepts'(all, P, GPs) :-
    '$get_all_all_percepts'(P, P2),
    filter_ground_all_percept__(P2, GPs).

%% get perepts list and filter out any terms that are not  wrapped
%% ground percept terms using updates interface
'$get_all_filtered_percepts'(updates, P, GPs) :-
    '$get_all_updates_percepts'(P, P2),
    filter_ground_updates_percept__(P2, GPs). 

'$log_process_percepts'(P) :-
    '$logging', !,
    open_string(write, StreamP),
    write(StreamP, 'Received new percepts:\n'),
    write(StreamP, P), nl(StreamP),
    send_log_message_(StreamP).
'$log_process_percepts'(_).

'$get_all_all_percepts'(_P, P2) :-
    ipc_peek(_, _, [timeout(poll)]), !,
    '$get_percepts'(P1),
    '$get_all_all_percepts'(P1, P2).
'$get_all_all_percepts'(P, P).

'$get_all_updates_percepts'(P, P2) :-
    ipc_peek(_, _, [timeout(poll)]), !,
    '$get_percepts'(P0),
    append(P, P0, P1),
    '$get_all_updates_percepts'(P1, P2).
'$get_all_updates_percepts'(P, P).



%%filter_ground_updates_percept__(A,B) :- errornl(filter_ground_updates_percept__(A,B)),fail.
filter_ground_updates_percept__([], []).
filter_ground_updates_percept__([r_(P)|Rest], [r_(P1)|Grest]) :-
    atom(P),
    P1 = P('$none_'),
    ground_percept_(P1), !,
    filter_ground_updates_percept__(Rest, Grest).
filter_ground_updates_percept__([r_(P)|Rest], [r_(P)|Grest]) :-
    ground_percept_(P), !,
    filter_ground_updates_percept__(Rest, Grest).
filter_ground_updates_percept__([u_(P)|Rest], [u_(P)|Grest]) :-
    ground_percept_(P), !,
    filter_ground_updates_percept__(Rest, Grest).
filter_ground_updates_percept__([f_(P)|Rest], [f_(P1)|Grest]) :-
    atom(P),
    P1 = P('$none_'),
    ground_percept_(P1), !,
    filter_ground_updates_percept__(Rest, Grest).
filter_ground_updates_percept__([f_(P)|Rest], [f_(P)|Grest]) :-
    percept_(P), !,
    filter_ground_updates_percept__(Rest, Grest).
filter_ground_updates_percept__([fa_(P)|Rest], [fa_(P1)|Grest]) :-
    atom(P),
    P1 = P('$none_'),
    ground_percept_(P1), !,
    filter_ground_updates_percept__(Rest, Grest).
filter_ground_updates_percept__([fa_(P)|Rest], [fa_(P)|Grest]) :-
    percept_(P), !,
    filter_ground_updates_percept__(Rest, Grest).
filter_ground_updates_percept__([_|Rest], Grest) :-
    filter_ground_updates_percept__(Rest, Grest).

filter_ground_all_percept__([], []).
filter_ground_all_percept__([P|Rest], [P1|Grest]) :-
    atom(P),
    P1 = P('$none_'),
    ground_percept_(P1), !,
    filter_ground_all_percept__(Rest, Grest).
filter_ground_all_percept__([P|Rest], [P|Grest]) :-
    ground_percept_(P), !,
    filter_ground_all_percept__(Rest, Grest).
filter_ground_all_percept__([_|Rest], Grest) :-
    filter_ground_all_percept__(Rest, Grest).


'$get_percepts'(Ps) :-
    Ps1 <<- _,
    (
      list(Ps1)
    ->
      Ps = Ps1
    ;
      Ps = [Ps1]
    ).
      


%% add resources hook if defined by user
'$assert_resources_hook_interface' :-
    current_predicate(resources_hook/1), !,
    assert((resources_callback :-  resources_hook(_))).
'$assert_resources_hook_interface' :-
    assert(resources_callback).

%% return the active (running) resources used
get_active_resources(Resources) :-
    findall(res(Task, Res),
            ('__running__'(Task),
                '__resources__'(Task,Res)),
            Resources1),
    (
     Resources1 = [all__]
    ->
     Resources = all
    ;
      Resources = Resources1
    ).

%% return resources being waited on
get_waiting_resources(Resources) :-
    findall(res(Task, Res),
            ('__waiting__'(Task),
                '__resources__'(Task,Res)),
            Resources1),
    (
     Resources1 = [all__]
    ->
     Resources = all
    ;
      Resources = Resources1
    ).


%% add interface to robotic messages if defined by user or used the
%% default handler
'$assert_send_robotic_message_interface' :-
    current_predicate(send_robotic_message/1), !,
    assert((send_robotic_message_interface_(A) :- send_robotic_message(A))).
'$assert_send_robotic_message_interface' :-
    assert((send_robotic_message_interface_(A) :- send_robotic_message_default_(A))).

%% add interface for post processing percepts
%% if defined by the user
'$assert_post_percepts_interface' :-
    current_predicate(post_process_percepts/2), !,
    assert((post_percepts_interface(A,B) :- post_process_percepts(A,B))).
'$assert_post_percepts_interface' :-
    assert((post_percepts_interface(_A,_B) :- true)).



controls_(C) :-
    send_robotic_message_interface_(C).

send_robotic_message_default_(Controls) :-
    '$agent_info'(_, embedded), !,
    Controls ->> embedded.
send_robotic_message_default_(Controls) :-
    '$agent_info'(_, Handle),
    to(Controls, Handle).


'$update_interfaces'(Interface) :-
    '$running_teleo', !,
    %% retract all interfaces
    '$retract_interface'(handle_percepts_interface_(_)),
    '$retract_interface'(handle_message_interface_(_, _)),
    '$retract_interface'(init_message_handler_interface),
    '$retract_interface'(init_percept_handler_interface),
    '$retract_interface'(overlapping_resources_interface_(_, _)),
    '$retract_interface'(resources_callback),
    '$retract_interface'(send_robotic_message_interface_(_)),
    '$retract_interface'(post_percepts_interface(_, _)),
    retractall('__percept_interface__'/1),
    
    %% set the percept handling interface (all or updates)
    '$assert_percepts_interface'(Interface),
    %% set the message handler for user defined handlers for:
    %% handle_message/2, handle_template_message/2, (handle_invalid_message/2
    %% if not defined do nothing
    '$assert_message_interface',
    %% set the handler to initialize message handler if defined by the user
    %% handler does nothing if not user defined
    '$assert_init_message_handler_interface',
    %% ditto for the percept handler
    '$assert_init_percept_handler_interface',
    %% set that handler for user defined overlapping_resources
    %% the handler fails if not user defined
    '$assert_overlapping_resources_interface',
    %% add resources hook if defined by user
    %% do nothing if not defined
    '$assert_resources_hook_interface',
    %% add interface for post processing percepts 
    %% if defined by the user - do nothing if not user defined
    '$assert_post_percepts_interface',
    %% add interface for sending robotic control messages
    '$assert_send_robotic_message_interface'.
'$update_interfaces'(_).

'$retract_interface'(Call) :-
    retractall(Call).

%% start_agent(Handle, Interface)
%% Start a teleor agent using Interface (all or updates) where
%% Handler is the address to send robot control messages to
start_agent(_Handle, _Interface) :-
    \+ pedro_is_registered,
     !,
    writeL_(["Error: Agent is not registered with pedro", nl_]).
start_agent(Handle, Interface) :-
    '$running_teleo', !,
    %%thread_handle((_:MyName@_)),
    %% agent messages from other agents are put in the messages threads buffer
    set_default_message_thread(messages),
    %% if the agent handle is an atom then localhost is added as the machine
    %% name otherwise Hndle will be treated as a local thread name
    ( atom(Handle) -> Handle1 = (Handle@localhost) ; Handle1 = Handle ),
    %% agent_info is used for sending messages to robot
    assert('$agent_info'(agent, Handle1)),
    %% set up user defined interfaces
    '$update_interfaces'(Interface),
    start_percept_thread__,
    thread_fork(messages, handle_messages_),
    thread_yield, %% to give the message thread a chance to init
    writeL_(["Messages thread started", nl_]),
    thread_fork(action_handler_, action_handler_),
    writeL_(["Actions thread started", nl_]),
    writeL_(["Agent initialised", nl_]).     
start_agent(_Handle, _Interface) :-
    writeL_(["Error: Attempting to start agent outside of the teleor interface",
	    nl_]).

%% start_embedded_agent(Reactivity) start an embedded agent using the all
%% interface. Reactivity is the time between polling perepts
start_embedded_agent(Reactivity) :-
    '$running_teleo', !,
    %%thread_handle((_:MyName@_)),
    set_default_message_thread(messages),
    assert('$agent_info'(agent, embedded)),
    '$update_interfaces'(all),
    %% start a small thread
    thread_fork(embedded, start_embedded_robot_control_,
                [choice_size(10),
                 env_size(10),
                 heap_size(50),
                 binding_trail_size(10),
                 other_trail_size(10),
                 scratchpad_size(10),
                 name_table_size(100),
                 ip_table_size(100)]),
    writeL_(["Starting embedded thread", nl_]),
    start_embedded_percept_thread__(Reactivity),
    thread_fork(messages, handle_messages_),
    thread_yield, %% to give the message a thread a chance to init
    writeL_(["Messages thread started", nl_]),
    thread_fork(action_handler_, action_handler_),
    writeL_(["Actions thread started", nl_]),
    writeL_(["Agent initialised", nl_]).
     
start_embedded_agent(_) :-
    writeL_(["Error: Attempting to start agent outside of the teleor interface",
	    nl_]).

start_embedded_percept_thread__(Reactivity) :-
    '$percept'(_), !,
    thread_fork(percepts,start_embedded_percept_handler_(Reactivity)),
    writeL_(["Starting percepts thread", nl_]),
    thread_wait_on_goal('$percept_thread_running').
start_embedded_percept_thread__(_Reactivity)  :-
    writeL_(stderr, ["Warning: the percept handling thread has not been started because there is no percept declaration", nl_]).

start_percept_thread__ :-
    '$percept'(_), !,
    thread_fork(percepts,start_percept_handler_),
    writeL_(["Starting percepts thread", nl_]),
    thread_wait_on_goal('$percept_thread_running').
start_percept_thread__ :-
    writeL_(stderr, ["Warning: the percept handling thread has not been started because there is no percept declaration", nl_]).


start_embedded_robot_control_ :-
    embedded_robot_control_loop_([]).

embedded_robot_control_loop_(OldActions) :- 
    Actions <<- _,
    control_device(OldActions, Actions),
    embedded_robot_control_loop_(Actions).

% updates_robot_thread_loop_(PrevPercepts, Reactivity) :-
%     (
%       ipc_peek(_, _, [timeout(Reactivity)])
%     ->
%       Actions <<- _ -> control_device(Actions)
%     ;
%       true
%     ),
%     poll_sensors(Percepts),
%     get_updates_percepts_(PrevPercepts, Percepts, UpdatesPercepts),
%     UpdatesPercepts ->> percepts,
%     updates_robot_thread_loop_(Percepts, Reactivity).
    
% all_robot_thread_loop_(Reactivity) :-
%     (
%       ipc_peek(_, _, [timeout(Reactivity)])
%     ->
%       Actions <<- _ -> control_device(Actions)
%     ;
%       true
%     ),
%     poll_sensors(Percepts),
%     Percepts ->> percepts,
%     all_robot_thread_loop_(Reactivity).

% get_updates_percepts_(OldPercepts, NewPercepts, Updates) :-
%     diff_list(OldPercepts, NewPercepts, ToForget),
%     diff_list(NewPercepts, OldPercepts, ToRemember),
%     map(add_f_, ToForget, Forget),
%     map(add_r_, ToRemember, Remember),
%     append(Forget, Remember, Updates).

% add_f_(F, f_(F)).
% add_r_(F, r_(F)).


% ensure_pedro_connected(Name, Host, Port) :-
%     \+ pedro_is_connected,
%     (
%       pedro_connect(Host, Port)
%     ->
%       true
%     ;
%       throw(cannot_connect_pedro('$none_'))
%     ),
%     (
%       pedro_register(Name)
%     ->
%       true
%     ;
%       throw(cannot_register_pedro(Name))
%     ).

% ensure_pedro_connected(MyName, _Host, _Port) :-
%     pedro_deregister,
%     (
%       pedro_register(MyName)
%     ->
%       writeL_(["Re-registering this process with Pedro using name: ", MyName,nl_])
%     ;
%       throw(cannot_register_pedro(MyName))
%     ).   

%% Kill the agent:
%% Send the message "finish" to the robot process
%% Kill all tasks
%% Exit "helper" threads
kill_agent(_) :-
    retract('$agent_info'(agent, Handle)), !,
    finish ->> Handle,
    forall(task(Name, _), kill_task(Name)),
    ( thread_is_thread(messages) -> thread_exit(messages) ; true ),
    thread_exit(percepts),
    thread_exit(action_handler_).
kill_agent(_) :-
    writeL_(["Error: No agent to kill", nl_]).



'$assert_percepts_interface'(all) :-
    !,
    assert('__percept_interface__'(all)),
    assert((handle_percepts_interface_(A) :- handle_percepts_all_(A, F, R), post_percepts_interface(F, R))).
'$assert_percepts_interface'(_) :-
    !,
    assert('__percept_interface__'(updates)),
    assert((handle_percepts_interface_(A) :- handle_percepts_updates_(A, F, R),post_percepts_interface(F, R))).

%%handle_percepts_all_(Ps, Forgets, Remembers) :- errornl(handle_percepts_all_(Ps, Forgets, Remembers)),fail.
handle_percepts_all_(Ps, Forgets, Remembers) :-
    handle_percepts_all_1_(Ps, Forgets, Remembers),
    %% Forgets are the current percepts that need to be forgotten
    %% Remebers are the new precepts that are not current and so need
    %% remembering
    '$forget_remember_percepts'(Forgets, Remembers).

handle_percepts_all_1_(Pers, Forgets, Remembers) :-
    findall(P, to_forget_percept_(P, Pers), Forgets),
    handle_percepts_all_aux_(Pers, Remembers).

to_forget_percept_(P, Pers) :-
    a_percept_(P),
    \+member(P, Pers).

%% Extract the new percepts - i.e. percepts that are in the percepts list
%% but are not current (so need to be remembered)
handle_percepts_all_aux_([], []).
handle_percepts_all_aux_([P|Ps], OutRemembers) :-
    call(P), !,
    OutRemembers = Remembers,
    handle_percepts_all_aux_(Ps, Remembers).
handle_percepts_all_aux_([P|Ps], OutRemembers) :-
    OutRemembers = [P|Remembers],
    handle_percepts_all_aux_(Ps, Remembers).

percept_(P) :- type(P, '??'(dyn_term)), '$percept_type_info'(_, _, P).
ground_percept_(P) :- type(P, '!'(dyn_term)), '$percept_type_info'(_, _, P).

%% See comment before strip_percept_patterns in consult.ql
%% Ps is a list consisting of terms of the form:
%% r_(P) - P is a ground percept that needs to be remembered
%% f_(P) - P is a percept that needs to be forgotten - if P
%%         contains variable the first matching current percept will
%%         be forgotten
%% fa_(P) - P is a percept that needs to be forgotten - typically P
%%          contains variables and all matching current percepts will
%%          be forgotten
%% u_(P) -  P is a ground percept - if it matches a current percept in
%%         non '^' tagged arguments then the matching percept will be forgotten
%%         and P remembered. OW P is remembered and nothing is forgotten
handle_percepts_updates_(Ps, Forgets3, Remembers4) :-
    handle_percepts_updates_aux_(Ps, Forgets1, Remembers1),
    %% Forgets1 and Remembers1 are the "raw" percepts to forget and remember
    %% but there can be repeats because these lists have been constructed
    %% from various percept update terms - the next 4 calls remove repeats
    diff_list(Forgets1, Remembers1, Forgets2),
    diff_list(Remembers1, Forgets1, Remembers2),
    remove_duplicates(Forgets2, Forgets3),
    remove_duplicates(Remembers2, Remembers3),
    handle_percepts_updates_aux_2__(Remembers3, Remembers4),
    %% Remembers4 are the percepts in Remembers3 that actually need
    %% to be remembered (not the ones that are already current)
    '$forget_remember_percepts'(Forgets3, Remembers4).

%%handle_percepts_updates_aux_(A,B,C,D) :- errornl(handle_percepts_updates_aux_(A,B,C,D)),fail.
handle_percepts_updates_aux_([], [], []).
handle_percepts_updates_aux_([u_(P)|Ps], [P1|Forgets], [P|Remembers]) :-
    percept_pattern_(P, P1),
    call(P1),
    %% P1 (a current percept) matches P in non '^' tagged arguments
    !, handle_percepts_updates_aux_(Ps, Forgets, Remembers).
handle_percepts_updates_aux_([u_(P)|Ps], Forgets, [P|Remembers]) :-
    !, handle_percepts_updates_aux_(Ps, Forgets, Remembers).
handle_percepts_updates_aux_([r_(P)|Ps], Forgets, [P|Remembers]) :-
    !, handle_percepts_updates_aux_(Ps, Forgets, Remembers).
handle_percepts_updates_aux_([f_(P)|Ps],  [P|Forgets], Remembers) :-
    call(P), !,
    handle_percepts_updates_aux_(Ps, Forgets, Remembers).
handle_percepts_updates_aux_([fa_(P)|Ps], AllForgets, Remembers) :-
    !,
    findall(P, call(P), FPs),
    append(FPs, Forgets, AllForgets),
    handle_percepts_updates_aux_(Ps, Forgets, Remembers).
handle_percepts_updates_aux_([_|Ps], Forgets, Remembers) :-
     handle_percepts_updates_aux_(Ps, Forgets, Remembers).

handle_percepts_updates_aux_2__([], []).
handle_percepts_updates_aux_2__([R|Rems], Remembers) :-
    call(R), !,
    handle_percepts_updates_aux_2__(Rems, Remembers).
handle_percepts_updates_aux_2__([R|Rems], [R|Remembers]) :-
    handle_percepts_updates_aux_2__(Rems, Remembers).

%%'$forget_remember_percepts'(Forgets, Remembers)
%% Forgets is the list of percepts to forget and
%% Remembers is the list of percepts to remember
'$forget_remember_percepts'([], []) :- !.
'$forget_remember_percepts'(Fs, _Rs) :-
    %% retract all the elements of Fs
    member(C, Fs),
    retract(C),
    fail.
'$forget_remember_percepts'(_Fs, Rs) :-
    %% asserta all elements of Rs
    member(C, Rs),
    asserta(C),
    fail.
'$forget_remember_percepts'(Fs, Rs) :-
    %% since the percepts have been updated we need to run infer_and_remember
    '$infer_and_remember'(Fs, Rs),
    '$reset_time_fact'.

%% on backtracking lists all percept facts in the BS
a_percept_(P) :-
    '$percept_type_info'(_, _, P),
    call(P).

%% start_task(Name, TR) - start a task with name Name with initial TR procedure
%% call TR 

%% If there are task_atomic declarations then the starting procedure must
%% be declared as a task_start procedure
start_task(Name, TR) :-
    '$task_atomic'(_),
    functor(TR, TRF, _),
    \+'$task_start'(TRF), !, 
    writeL_(stderr, ["Error: for " , start_task(Name, TR),
                    " ", TR,
                    " has not been declared as a task start procedure", nl_]).

%% If the name is not given then the name is instantiated to the thread name
%% whose name has root name being the functor of TR
start_task(Name, TR) :-
    var(Name),!,
    thread_wait_on_goal('$percept_thread_running'),
    functor(TR, TRF, _N),
    thread_fork(Task,  start_task_aux_(TR), TRF),
    Name = Task,
    assert(task(Task,TR)),
    writeL_(["Starting task ", Task, nl_]),
    '$forget_remember_aux'(['__resources__'(Task, _)],
                           ['__resources__'(Task, [])]).


start_task(Name, TR) :-
    thread_wait_on_goal('$percept_thread_running'),
    thread_fork(_, '$start_task'(Name, TR)),
    thread_wait_on_goal(task(Name, _)).

'$start_task'(Name, TR) :-
    (
      thread_is_thread(Name)
    ->
      thread_sleep(1),
      (
        thread_is_thread(Name)
      ->
        writeL_(stderr, ["Error: For " , start_task(Name, TR),
                        " the name is already in use", nl_]),
        thread_exit
      ;
        true
      )
    ;
      true
    ),
    thread_set_symbol(Name),
    writeL_(["Starting task ", Name, nl_]),
    '$forget_remember_aux'(['__resources__'(Name, _)],
                           ['__resources__'(Name, [])]),
    assert(task(Name,TR)),
    start_task_aux_(TR).

%% kill the task with name Name
kill_task(Name) :-
    thread_is_thread(Name), !,
    writeL_(stderr, ["Exiting ", Name, nl_]),
    retract(task(Name,_)),
    assert('$should_kill'(Name)),
    '$reset_time_fact',
    thread_yield.
kill_task(Name) :-
        writeL_(stderr, ["Error: For " , kill_task(Name),
                        " the task is not running", nl]).

%% Run in the messages thread to handle messages
handle_messages_ :-
    %% possible user defined initialization - e.g. subscribe
    thread_atomic_goal(init_message_handler_interface),
    '$handle_messages'.

'$handle_messages' :-
    repeat,
    from(M, Addr, true),
    %% there are messages to process - wait until task re-eval finished 
    thread_wait_on_goal('$can_run_handlers'),
    thread_atomic_goal('$process_messages'(M, Addr)),
    thread_yield,
    fail.

'$process_messages'(M, Addr) :-
    '$process_messages_aux'(M, Addr),
    '$eval_tasks'.

'$process_messages_aux'(M, Addr) :-
    (
      M == quiting
    ->
      thread_exit(thread0)
    ;
      %% process messages
      thread_atomic_goal(handle_message_interface_(M, Addr))
    ),
    (
      ipc_peek(_, _, [timeout(poll)])
    ->
      from(M1, Addr1, true),
      %% keep going if more have turned up
      '$process_messages_aux'(M1, Addr1)
    ;
      true
    ).

'$set_can_run_handlers' :-
    '$can_run_handlers', !.
'$set_can_run_handlers' :-
    assert('$can_run_handlers').

action_handler_ :-
    repeat,
    Actions <<- _,
    %% there are actions to process - wait until task re-eval finished 
    thread_wait_on_goal('$can_run_handlers'),
    thread_atomic_goal('$process_actions'(Actions)),
    thread_yield,
    fail.

'$process_actions'(Actions) :-
    '$process_actions_aux'(Actions),
    '$eval_tasks'.


'$process_actions_aux'(Actions) :-
    %% process all the actions
    forall(member(A, Actions), '$call_tr_call'(A)),
    (
      ipc_peek(_, _, [timeout(poll)])
    ->
      Actions1 <<- _,
      %% keep going if more have turned up
      '$process_actions_aux'(Actions1)
    ;
      true
    ).


'$call_tr_call'(G) :- call(G), !.
'$call_tr_call'(G) :- writeL_(stderr, ["The call ", G, " failed", nl_]).
    


'$assert_message_interface' :-
    (
      current_predicate(handle_message/2)
    ;
      current_predicate(handle_template_message/2)
    ),
    !,
    (
      current_predicate(handle_message/2),
      \+ '$user_type_info'(message, defined_type, _, _, _)
    ->
      '$display_error'(message_type_undefined)
    ;
      current_predicate(handle_message/2)
    ->
      assert((handle_message_interface_(M, A) :-
             type(M, '!'(message)), !,
              log_message_(message(M, A)),
              handle_message(M, A)))
    ;
      true
    ),
    (
      current_predicate(handle_template_message/2),
      \+ '$user_type_info'(template_message, defined_type, _, _, _)
    ->
      '$display_error'(template_message_type_undefined)
    ;
      current_predicate(handle_template_message/2)
    ->
      assert((handle_message_interface_(M, A) :-
             type(M, '?'(message)),
              template(M), !,
             log_message_(message(M, A)),
              handle_template_message(M, A)))
    ;
      true
    ),
    
    (
      current_predicate(handle_invalid_message/2)
    ->
      assert((handle_message_interface_(M, A) :-
             log_message_(message(M, A)),
             handle_invalid_message_(M, A)))
    ;
      assert(handle_message_interface_(_M, _A))
    ).
'$assert_message_interface' :-
    assert((handle_message_interface_(M, A) :-
           log_message_(message(M, A)),
	   '$default_handle_message'(M,A))).



'$assert_init_message_handler_interface' :-
    current_predicate(init_message_handler/1), !,
    assert((init_message_handler_interface :- init_message_handler(_))).
'$assert_init_message_handler_interface' :-
    assert(init_message_handler_interface).

'$assert_init_percept_handler_interface' :-
    current_predicate(init_percept_handler/1), !,
    assert((init_percept_handler_interface :- init_percept_handler(_))).
'$assert_init_percept_handler_interface' :-
    assert(init_percept_handler_interface).


'$assert_overlapping_resources_interface' :-
    current_predicate(overlapping_resources/2), !,
    assert((overlapping_resources_interface_(A, B) :-
           overlapping_resources(A,B))).
'$assert_overlapping_resources_interface' :-
    assert((overlapping_resources_interface_(_, _) :- fail)).

'$default_handle_message'(M,A) :-
    ground(M), !,
    '__time__'(Time),
    remember(received_message(M, A, Time)).
'$default_handle_message'(_, _).
	    
